@@ -48,9 +48,20 @@ typedef struct
 typedef enum
 {
   MONITOR_MATCH_ALL = 0,
-  MONITOR_MATCH_EXTERNAL = (1 << 0)
+  MONITOR_MATCH_EXTERNAL = (1 << 0),
+  MONITOR_MATCH_BUILTIN = (1 << 1),
+  MONITOR_MATCH_VISIBLE = (1 << 2),
+  MONITOR_MATCH_WITH_SUGGESTED_POSITION = (1 << 3),
+  MONITOR_MATCH_PRIMARY = (1 << 4),
+  MONITOR_MATCH_ALLOW_FALLBACK = (1 << 5)
 } MonitorMatchRule;
 
+typedef enum
+{
+  MONITOR_POSITIONING_LINEAR,
+  MONITOR_POSITIONING_SUGGESTED,
+} MonitorPositioningMode;
+
 struct _GfMonitorConfigManager
 {
   GObject               parent;
@@ -65,6 +76,22 @@ struct _GfMonitorConfigManager
 
 G_DEFINE_TYPE (GfMonitorConfigManager, gf_monitor_config_manager, G_TYPE_OBJECT)
 
+static GfMonitorsConfig *
+get_root_config (GfMonitorsConfig *config)
+{
+  if (config->parent_config == NULL)
+    return config;
+
+  return get_root_config (config->parent_config);
+}
+
+static gboolean
+has_same_root_config (GfMonitorsConfig *config_a,
+                      GfMonitorsConfig *config_b)
+{
+  return get_root_config (config_a) == get_root_config (config_b);
+}
+
 static void
 history_unref (gpointer data,
                gpointer user_data)
@@ -72,6 +99,77 @@ history_unref (gpointer data,
   g_object_unref (data);
 }
 
+static gboolean
+is_lid_closed (GfMonitorManager *monitor_manager)
+{
+  GfBackend *backend;
+
+  backend = gf_monitor_manager_get_backend (monitor_manager);
+
+  return gf_backend_is_lid_closed (backend);
+}
+
+static gboolean
+monitor_matches_rule (GfMonitor        *monitor,
+                      GfMonitorManager *monitor_manager,
+                      MonitorMatchRule  match_rule)
+{
+  if (monitor == NULL)
+    return FALSE;
+
+  if (match_rule & MONITOR_MATCH_BUILTIN)
+    {
+      if (!gf_monitor_is_laptop_panel (monitor))
+        return FALSE;
+    }
+  else if (match_rule & MONITOR_MATCH_EXTERNAL)
+    {
+      if (gf_monitor_is_laptop_panel (monitor))
+        return FALSE;
+    }
+
+  if (match_rule & MONITOR_MATCH_VISIBLE)
+    {
+      if (gf_monitor_is_laptop_panel (monitor) &&
+          is_lid_closed (monitor_manager))
+        return FALSE;
+    }
+
+  if (match_rule & MONITOR_MATCH_WITH_SUGGESTED_POSITION)
+    {
+      if (!gf_monitor_get_suggested_position (monitor, NULL, NULL))
+        return FALSE;
+    }
+
+  return TRUE;
+}
+
+static GList *
+find_monitors (GfMonitorManager *monitor_manager,
+               MonitorMatchRule  match_rule,
+               GfMonitor        *not_this_one)
+{
+  GList *result;
+  GList *monitors;
+  GList *l;
+
+  result = NULL;
+  monitors = gf_monitor_manager_get_monitors (monitor_manager);
+
+  for (l = g_list_last (monitors); l; l = l->prev)
+    {
+      GfMonitor *monitor = l->data;
+
+      if (not_this_one != NULL && monitor == not_this_one)
+        continue;
+
+      if (monitor_matches_rule (monitor, monitor_manager, match_rule))
+        result = g_list_prepend (result, monitor);
+    }
+
+  return result;
+}
+
 static GfMonitor *
 find_monitor_with_highest_preferred_resolution (GfMonitorManager *monitor_manager,
                                                 MonitorMatchRule  match_rule)
@@ -81,7 +179,8 @@ find_monitor_with_highest_preferred_resolution (GfMonitorManager *monitor_manage
   int largest_area = 0;
   GfMonitor *largest_monitor = NULL;
 
-  monitors = gf_monitor_manager_get_monitors (monitor_manager);
+  monitors = find_monitors (monitor_manager, match_rule, NULL);
+
   for (l = monitors; l; l = l->next)
     {
       GfMonitor *monitor = l->data;
@@ -89,12 +188,6 @@ find_monitor_with_highest_preferred_resolution (GfMonitorManager *monitor_manage
       int width, height;
       int area;
 
-      if (match_rule & MONITOR_MATCH_EXTERNAL)
-        {
-          if (gf_monitor_is_laptop_panel (monitor))
-            continue;
-        }
-
       mode = gf_monitor_get_preferred_mode (monitor);
       gf_monitor_mode_get_resolution (mode, &width, &height);
       area = width * height;
@@ -106,6 +199,8 @@ find_monitor_with_highest_preferred_resolution (GfMonitorManager *monitor_manage
         }
     }
 
+  g_clear_pointer (&monitors, g_list_free);
+
   return largest_monitor;
 }
 
@@ -120,38 +215,32 @@ find_monitor_with_highest_preferred_resolution (GfMonitorManager *monitor_manage
  * alternatives, except if no other alternatives exist.
  */
 static GfMonitor *
-find_primary_monitor (GfMonitorManager *monitor_manager)
+find_primary_monitor (GfMonitorManager *monitor_manager,
+                      MonitorMatchRule  match_rule)
 {
   GfMonitor *monitor;
 
-  if (gf_monitor_manager_is_lid_closed (monitor_manager))
-    {
-      monitor = gf_monitor_manager_get_primary_monitor (monitor_manager);
-      if (monitor && !gf_monitor_is_laptop_panel (monitor))
-        return monitor;
+  monitor = gf_monitor_manager_get_primary_monitor (monitor_manager);
 
-      monitor = find_monitor_with_highest_preferred_resolution (monitor_manager,
-                                                                MONITOR_MATCH_EXTERNAL);
+  if (monitor_matches_rule (monitor, monitor_manager, match_rule))
+    return monitor;
 
-      if (monitor)
-        return monitor;
+  monitor = gf_monitor_manager_get_laptop_panel (monitor_manager);
 
-      return find_monitor_with_highest_preferred_resolution (monitor_manager,
-                                                             MONITOR_MATCH_ALL);
-    }
-  else
-    {
-      monitor = gf_monitor_manager_get_primary_monitor (monitor_manager);
-      if (monitor)
-        return monitor;
+  if (monitor_matches_rule (monitor, monitor_manager, match_rule))
+    return monitor;
 
-      monitor = gf_monitor_manager_get_laptop_panel (monitor_manager);
-      if (monitor)
-        return monitor;
+  monitor = find_monitor_with_highest_preferred_resolution (monitor_manager,
+                                                            match_rule);
 
-      return find_monitor_with_highest_preferred_resolution (monitor_manager,
-                                                             MONITOR_MATCH_ALL);
-    }
+  if (monitor != NULL)
+    return monitor;
+
+  if (match_rule & MONITOR_MATCH_ALLOW_FALLBACK)
+    return find_monitor_with_highest_preferred_resolution (monitor_manager,
+                                                           MONITOR_MATCH_ALL);
+
+  return NULL;
 }
 
 static GfMonitorTransform
@@ -160,29 +249,17 @@ get_monitor_transform (GfMonitorManager *monitor_manager,
 {
   GfBackend *backend;
   GfOrientationManager *orientation_manager;
+  GfOrientation orientation;
 
-  if (!gf_monitor_is_laptop_panel (monitor))
+  if (!gf_monitor_is_laptop_panel (monitor) ||
+      !gf_monitor_manager_get_panel_orientation_managed (monitor_manager))
     return GF_MONITOR_TRANSFORM_NORMAL;
 
   backend = gf_monitor_manager_get_backend (monitor_manager);
   orientation_manager = gf_backend_get_orientation_manager (backend);
+  orientation = gf_orientation_manager_get_orientation (orientation_manager);
 
-  switch (gf_orientation_manager_get_orientation (orientation_manager))
-    {
-      case GF_ORIENTATION_BOTTOM_UP:
-        return GF_MONITOR_TRANSFORM_180;
-
-      case GF_ORIENTATION_LEFT_UP:
-        return GF_MONITOR_TRANSFORM_90;
-
-      case GF_ORIENTATION_RIGHT_UP:
-        return GF_MONITOR_TRANSFORM_270;
-
-      case GF_ORIENTATION_UNDEFINED:
-      case GF_ORIENTATION_NORMAL:
-      default:
-        return GF_MONITOR_TRANSFORM_NORMAL;
-    }
+  return gf_monitor_transform_from_orientation (orientation);
 }
 
 static GfLogicalMonitorConfig *
@@ -190,12 +267,11 @@ create_preferred_logical_monitor_config (GfMonitorManager           *monitor_man
                                          GfMonitor                  *monitor,
                                          int                         x,
                                          int                         y,
-                                         GfLogicalMonitorConfig     *primary_logical_monitor_config,
+                                         float                       scale,
                                          GfLogicalMonitorLayoutMode  layout_mode)
 {
   GfMonitorMode *mode;
   int width, height;
-  float scale;
   GfMonitorTransform transform;
   GfMonitorConfig *monitor_config;
   GfLogicalMonitorConfig *logical_monitor_config;
@@ -203,14 +279,6 @@ create_preferred_logical_monitor_config (GfMonitorManager           *monitor_man
   mode = gf_monitor_get_preferred_mode (monitor);
   gf_monitor_mode_get_resolution (mode, &width, &height);
 
-  if ((gf_monitor_manager_get_capabilities (monitor_manager) &
-       GF_MONITOR_MANAGER_CAPABILITY_GLOBAL_SCALE_REQUIRED) &&
-      primary_logical_monitor_config)
-    scale = primary_logical_monitor_config->scale;
-  else
-    scale = gf_monitor_manager_calculate_monitor_mode_scale (monitor_manager,
-                                                             monitor, mode);
-
   switch (layout_mode)
     {
       case GF_LOGICAL_MONITOR_LAYOUT_MODE_LOGICAL:
@@ -266,6 +334,7 @@ create_for_switch_config_all_mirror (GfMonitorConfigManager *config_manager)
   GList *l;
   GfMonitorsConfig *monitors_config;
 
+  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
   monitors = gf_monitor_manager_get_monitors (monitor_manager);
   monitor = monitors->data;
   modes = gf_monitor_get_modes (monitor);
@@ -337,7 +406,11 @@ create_for_switch_config_all_mirror (GfMonitorConfigManager *config_manager)
       if (!mode)
         continue;
 
-      scale = gf_monitor_manager_calculate_monitor_mode_scale (monitor_manager, l_monitor, mode);
+      scale = gf_monitor_manager_calculate_monitor_mode_scale (monitor_manager,
+                                                               layout_mode,
+                                                               l_monitor,
+                                                               mode);
+
       best_scale = MAX (best_scale, scale);
       monitor_configs = g_list_prepend (monitor_configs, gf_monitor_config_new (l_monitor, mode));
     }
@@ -355,7 +428,6 @@ create_for_switch_config_all_mirror (GfMonitorConfigManager *config_manager)
   };
 
   logical_monitor_configs = g_list_append (NULL, logical_monitor_config);
-  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
 
   monitors_config = gf_monitors_config_new (monitor_manager,
                                             logical_monitor_configs,
@@ -369,86 +441,211 @@ create_for_switch_config_all_mirror (GfMonitorConfigManager *config_manager)
   return monitors_config;
 }
 
-static GfMonitorsConfig *
-create_for_switch_config_external (GfMonitorConfigManager *config_manager)
+static gboolean
+verify_suggested_monitors_config (GList *logical_monitor_configs)
 {
-  GfMonitorManager *monitor_manager = config_manager->monitor_manager;
-  GList *logical_monitor_configs = NULL;
-  int x = 0;
-  GfLogicalMonitorLayoutMode layout_mode;
-  GList *monitors;
+  GList *region;
   GList *l;
-  GfMonitorsConfig *monitors_config;
+
+  region = NULL;
+
+  for (l = logical_monitor_configs; l; l = l->next)
+    {
+      GfLogicalMonitorConfig *logical_monitor_config = l->data;
+      GfRectangle *rect = &logical_monitor_config->layout;
+
+      if (gf_rectangle_overlaps_with_region (region, rect))
+        {
+          g_warning ("Suggested monitor config has overlapping region, "
+                      "rejecting");
+
+          g_list_free (region);
+
+          return FALSE;
+        }
+
+      region = g_list_prepend (region, rect);
+    }
+
+  for (l = region; region->next && l; l = l->next)
+    {
+      GfRectangle *rect = l->data;
+
+      if (!gf_rectangle_is_adjacent_to_any_in_region (region, rect))
+        {
+          g_warning ("Suggested monitor config has monitors with no "
+                      "neighbors, rejecting");
+
+          g_list_free (region);
+
+          return FALSE;
+        }
+    }
+
+  g_list_free (region);
+
+  return TRUE;
+}
+
+static float
+compute_scale_for_monitor (GfMonitorManager *monitor_manager,
+                           GfMonitor        *monitor,
+                           GfMonitor        *primary_monitor)
+
+{
+  GfMonitor *target_monitor;
+  GfMonitorManagerCapability capabilities;
+  GfLogicalMonitorLayoutMode layout_mode;
+  GfMonitorMode *monitor_mode;
+
+  target_monitor = monitor;
+  capabilities = gf_monitor_manager_get_capabilities (monitor_manager);
+
+  if ((capabilities & GF_MONITOR_MANAGER_CAPABILITY_GLOBAL_SCALE_REQUIRED) &&
+      primary_monitor != NULL)
+    {
+      target_monitor = primary_monitor;
+    }
 
   layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
+  monitor_mode = gf_monitor_get_preferred_mode (target_monitor);
+
+  return gf_monitor_manager_calculate_monitor_mode_scale (monitor_manager,
+                                                          layout_mode,
+                                                          target_monitor,
+                                                          monitor_mode);
+}
+
+static GfMonitorsConfig *
+create_monitors_config (GfMonitorConfigManager *config_manager,
+                        MonitorMatchRule        match_rule,
+                        MonitorPositioningMode  positioning,
+                        GfMonitorsConfigFlag    config_flags)
+{
+  GfMonitorManager *monitor_manager = config_manager->monitor_manager;
+  GfMonitor *primary_monitor;
+  GfLogicalMonitorLayoutMode layout_mode;
+  GList *logical_monitor_configs;
+  float scale;
+  int x, y;
+  GList *monitors;
+  GList *l;
+
+  primary_monitor = find_primary_monitor (monitor_manager,
+                                          match_rule | MONITOR_MATCH_VISIBLE);
+
+  if (!primary_monitor)
+    return NULL;
+
+  x = y = 0;
+  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
+
+  monitors = NULL;
+  if (!(match_rule & MONITOR_MATCH_PRIMARY))
+    monitors = find_monitors (monitor_manager, match_rule, primary_monitor);
+
+  /*
+   * The primary monitor needs to be at the head of the list for the
+   * linear positioning to be correct.
+   */
+  monitors = g_list_prepend (monitors, primary_monitor);
 
-  monitors = gf_monitor_manager_get_monitors (monitor_manager);
   for (l = monitors; l; l = l->next)
     {
       GfMonitor *monitor = l->data;
       GfLogicalMonitorConfig *logical_monitor_config;
+      gboolean has_suggested_position;
 
-      if (gf_monitor_is_laptop_panel (monitor))
-        continue;
+      switch (positioning)
+        {
+          case MONITOR_POSITIONING_SUGGESTED:
+            has_suggested_position = gf_monitor_get_suggested_position (monitor,
+                                                                        &x,
+                                                                        &y);
+            g_assert (has_suggested_position);
+            break;
+
+          case MONITOR_POSITIONING_LINEAR:
+          default:
+            break;
+        }
+
+      scale = compute_scale_for_monitor (monitor_manager,
+                                         monitor,
+                                         primary_monitor);
 
       logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                        monitor, x, 0, NULL,
+                                                                        monitor,
+                                                                        x,
+                                                                        y,
+                                                                        scale,
                                                                         layout_mode);
 
-      logical_monitor_configs = g_list_append (logical_monitor_configs,
-                                               logical_monitor_config);
-
-      if (x == 0)
-        logical_monitor_config->is_primary = TRUE;
+      logical_monitor_config->is_primary = (monitor == primary_monitor);
+      logical_monitor_configs = g_list_append (logical_monitor_configs, logical_monitor_config);
 
       x += logical_monitor_config->layout.width;
     }
 
-  monitors_config = gf_monitors_config_new (monitor_manager,
-                                            logical_monitor_configs,
-                                            layout_mode,
-                                            GF_MONITORS_CONFIG_FLAG_NONE);
+  g_clear_pointer (&monitors, g_list_free);
 
-  if (monitors_config)
-    gf_monitors_config_set_switch_config (monitors_config,
-                                          GF_MONITOR_SWITCH_CONFIG_EXTERNAL);
+  if (positioning == MONITOR_POSITIONING_SUGGESTED)
+    {
+      if (!verify_suggested_monitors_config (logical_monitor_configs))
+        {
+          g_list_free_full (logical_monitor_configs,
+                            (GDestroyNotify) gf_logical_monitor_config_free);
+
+          return NULL;
+        }
+    }
+
+  return gf_monitors_config_new (monitor_manager,
+                                 logical_monitor_configs,
+                                 layout_mode,
+                                 config_flags);
+}
+
+static GfMonitorsConfig *
+create_monitors_switch_config (GfMonitorConfigManager    *config_manager,
+                               MonitorMatchRule           match_rule,
+                               MonitorPositioningMode     positioning,
+                               GfMonitorsConfigFlag       config_flags,
+                               GfMonitorSwitchConfigType  switch_config)
+{
+  GfMonitorsConfig *monitors_config;
+
+  monitors_config = create_monitors_config (config_manager,
+                                            match_rule,
+                                            positioning,
+                                            config_flags);
+
+  if (monitors_config == NULL)
+    return NULL;
+
+  gf_monitors_config_set_switch_config (monitors_config, switch_config);
 
   return monitors_config;
 }
 
+static GfMonitorsConfig *
+create_for_switch_config_external (GfMonitorConfigManager *config_manager)
+{
+  return create_monitors_switch_config (config_manager,
+                                        MONITOR_MATCH_EXTERNAL,
+                                        MONITOR_POSITIONING_LINEAR,
+                                        GF_MONITORS_CONFIG_FLAG_NONE,
+                                        GF_MONITOR_SWITCH_CONFIG_EXTERNAL);
+}
+
 static GfMonitorsConfig *
 create_for_switch_config_builtin (GfMonitorConfigManager *config_manager)
 {
-  GfMonitorManager *monitor_manager = config_manager->monitor_manager;
-  GfLogicalMonitorLayoutMode layout_mode;
-  GList *logical_monitor_configs;
-  GfLogicalMonitorConfig *primary_logical_monitor_config;
-  GfMonitor *monitor;
-  GfMonitorsConfig *monitors_config;
-
-  monitor = gf_monitor_manager_get_laptop_panel (monitor_manager);
-  if (!monitor)
-    return NULL;
-
-  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
-
-  primary_logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                            monitor, 0, 0, NULL,
-                                                                            layout_mode);
-
-  primary_logical_monitor_config->is_primary = TRUE;
-  logical_monitor_configs = g_list_append (NULL, primary_logical_monitor_config);
-
-  monitors_config = gf_monitors_config_new (monitor_manager,
-                                            logical_monitor_configs,
-                                            layout_mode,
-                                            GF_MONITORS_CONFIG_FLAG_NONE);
-
-  if (monitors_config)
-    gf_monitors_config_set_switch_config (monitors_config,
-                                          GF_MONITOR_SWITCH_CONFIG_BUILTIN);
-
-  return monitors_config;
+  return create_monitors_switch_config (config_manager,
+                                        MONITOR_MATCH_BUILTIN,
+                                        MONITOR_POSITIONING_LINEAR,
+                                        GF_MONITORS_CONFIG_FLAG_NONE,
+                                        GF_MONITOR_SWITCH_CONFIG_BUILTIN);
 }
 
 static GList *
@@ -507,15 +704,15 @@ clone_logical_monitor_config_list (GList *configs_in)
 }
 
 static GfLogicalMonitorConfig *
-find_logical_config_for_builtin_display_rotation (GfMonitorConfigManager *config_manager,
-                                                  GList                  *logical_monitor_configs)
+find_logical_config_for_builtin_monitor (GfMonitorConfigManager *config_manager,
+                                         GList                  *logical_monitor_configs)
 {
   GfMonitor *panel;
   GList *l;
 
   panel = gf_monitor_manager_get_laptop_panel (config_manager->monitor_manager);
 
-  if (panel == NULL || !gf_monitor_is_active (panel))
+  if (panel == NULL)
     return NULL;
 
   for (l = logical_monitor_configs; l; l = l->next)
@@ -535,7 +732,15 @@ find_logical_config_for_builtin_display_rotation (GfMonitorConfigManager *config
       monitor_config = logical_monitor_config->monitor_configs->data;
       if (gf_monitor_spec_equals (gf_monitor_get_spec (panel),
                                   monitor_config->monitor_spec))
-        return logical_monitor_config;
+        {
+          GfMonitorMode *mode;
+
+          mode = gf_monitor_get_mode_from_spec (panel,
+                                                monitor_config->mode_spec);
+
+          if (mode != NULL)
+            return logical_monitor_config;
+        }
     }
 
   return NULL;
@@ -544,23 +749,24 @@ find_logical_config_for_builtin_display_rotation (GfMonitorConfigManager *config
 
 static GfMonitorsConfig *
 create_for_builtin_display_rotation (GfMonitorConfigManager *config_manager,
+                                     GfMonitorsConfig       *base_config,
                                      gboolean                rotate,
                                      GfMonitorTransform      transform)
 {
   GfMonitorManager *monitor_manager = config_manager->monitor_manager;
   GfLogicalMonitorConfig *logical_monitor_config;
   GfLogicalMonitorConfig *current_logical_monitor_config;
+  GfMonitorsConfig *config;
   GList *logical_monitor_configs;
   GList *current_configs;
   GfLogicalMonitorLayoutMode layout_mode;
   GList *current_monitor_configs;
 
-  if (!config_manager->current_config)
-    return NULL;
+  g_return_val_if_fail (base_config, NULL);
 
-  current_configs = config_manager->current_config->logical_monitor_configs;
-  current_logical_monitor_config = find_logical_config_for_builtin_display_rotation (config_manager,
-                                                                                     current_configs);
+  current_configs = base_config->logical_monitor_configs;
+  current_logical_monitor_config = find_logical_config_for_builtin_monitor (config_manager,
+                                                                            current_configs);
 
   if (!current_logical_monitor_config)
     return NULL;
@@ -585,11 +791,11 @@ create_for_builtin_display_rotation (GfMonitorConfigManager *config_manager,
   if (current_logical_monitor_config->transform == transform)
     return NULL;
 
-  current_monitor_configs = config_manager->current_config->logical_monitor_configs;
+  current_monitor_configs = base_config->logical_monitor_configs;
   logical_monitor_configs = clone_logical_monitor_config_list (current_monitor_configs);
 
-  logical_monitor_config = find_logical_config_for_builtin_display_rotation (config_manager,
-                                                                             logical_monitor_configs);
+  logical_monitor_config = find_logical_config_for_builtin_monitor (config_manager,
+                                                                    logical_monitor_configs);
   logical_monitor_config->transform = transform;
 
   if (gf_monitor_transform_is_rotated (current_logical_monitor_config->transform) !=
@@ -601,10 +807,16 @@ create_for_builtin_display_rotation (GfMonitorConfigManager *config_manager,
       logical_monitor_config->layout.height = temp;
     }
 
-  layout_mode = config_manager->current_config->layout_mode;
+  layout_mode = base_config->layout_mode;
 
-  return gf_monitors_config_new (monitor_manager, logical_monitor_configs,
-                                 layout_mode, GF_MONITORS_CONFIG_FLAG_NONE);
+  config = gf_monitors_config_new (monitor_manager,
+                                   logical_monitor_configs,
+                                   layout_mode,
+                                   GF_MONITORS_CONFIG_FLAG_NONE);
+
+  gf_monitors_config_set_parent_config (config, base_config);
+
+  return config;
 }
 
 static gboolean
@@ -922,7 +1134,7 @@ gf_create_monitors_config_key_for_current_state (GfMonitorManager *monitor_manag
         {
           laptop_monitor_spec = gf_monitor_get_spec (monitor);
 
-          if (gf_monitor_manager_is_lid_closed (monitor_manager))
+          if (is_lid_closed (monitor_manager))
             continue;
         }
 
@@ -1118,177 +1330,76 @@ gf_monitor_config_manager_get_stored (GfMonitorConfigManager *config_manager)
 GfMonitorsConfig *
 gf_monitor_config_manager_create_linear (GfMonitorConfigManager *config_manager)
 {
-  GfMonitorManager *monitor_manager = config_manager->monitor_manager;
-  GList *logical_monitor_configs;
-  GfMonitor *primary_monitor;
-  GfLogicalMonitorLayoutMode layout_mode;
-  GfLogicalMonitorConfig *primary_logical_monitor_config;
-  int x;
-  GList *monitors;
-  GList *l;
-  GfMonitorsConfig *monitors_config;
-
-  primary_monitor = find_primary_monitor (monitor_manager);
-  if (!primary_monitor)
-    return NULL;
-
-  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
-
-  primary_logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                            primary_monitor,
-                                                                            0, 0, NULL,
-                                                                            layout_mode);
-
-  primary_logical_monitor_config->is_primary = TRUE;
-  logical_monitor_configs = g_list_append (NULL, primary_logical_monitor_config);
-
-  x = primary_logical_monitor_config->layout.width;
-  monitors = gf_monitor_manager_get_monitors (monitor_manager);
-  for (l = monitors; l; l = l->next)
-    {
-      GfMonitor *monitor = l->data;
-      GfLogicalMonitorConfig *logical_monitor_config;
-
-      if (monitor == primary_monitor)
-        continue;
-
-      if (gf_monitor_is_laptop_panel (monitor) &&
-          gf_monitor_manager_is_lid_closed (monitor_manager))
-        continue;
-
-      logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                        monitor, x, 0,
-                                                                        primary_logical_monitor_config,
-                                                                        layout_mode);
-
-      logical_monitor_configs = g_list_append (logical_monitor_configs,
-                                               logical_monitor_config);
-
-      x += logical_monitor_config->layout.width;
-    }
-
-  monitors_config = gf_monitors_config_new (monitor_manager,
-                                            logical_monitor_configs,
-                                            layout_mode,
-                                            GF_MONITORS_CONFIG_FLAG_NONE);
-
-  if (monitors_config)
-    gf_monitors_config_set_switch_config (monitors_config,
-                                          GF_MONITOR_SWITCH_CONFIG_ALL_LINEAR);
-
-  return monitors_config;
+  return create_monitors_config (config_manager,
+                                 MONITOR_MATCH_VISIBLE |
+                                 MONITOR_MATCH_ALLOW_FALLBACK,
+                                 MONITOR_POSITIONING_LINEAR,
+                                 GF_MONITORS_CONFIG_FLAG_NONE);
 }
 
 GfMonitorsConfig *
 gf_monitor_config_manager_create_fallback (GfMonitorConfigManager *config_manager)
 {
-  GfMonitorManager *monitor_manager = config_manager->monitor_manager;
-  GfMonitor *primary_monitor;
-  GList *logical_monitor_configs;
-  GfLogicalMonitorLayoutMode layout_mode;
-  GfLogicalMonitorConfig *primary_logical_monitor_config;
-
-  primary_monitor = find_primary_monitor (monitor_manager);
-  if (!primary_monitor)
-    return NULL;
-
-  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
-
-  primary_logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                            primary_monitor,
-                                                                            0, 0, NULL,
-                                                                            layout_mode);
-
-  primary_logical_monitor_config->is_primary = TRUE;
-  logical_monitor_configs = g_list_append (NULL, primary_logical_monitor_config);
-
-  return gf_monitors_config_new (monitor_manager, logical_monitor_configs,
-                                 layout_mode, GF_MONITORS_CONFIG_FLAG_NONE);
+  return create_monitors_config (config_manager,
+                                 MONITOR_MATCH_PRIMARY |
+                                 MONITOR_MATCH_ALLOW_FALLBACK,
+                                 MONITOR_POSITIONING_LINEAR,
+                                 GF_MONITORS_CONFIG_FLAG_NONE);
 }
 
 GfMonitorsConfig *
 gf_monitor_config_manager_create_suggested (GfMonitorConfigManager *config_manager)
 {
-  GfMonitorManager *monitor_manager = config_manager->monitor_manager;
-  GfLogicalMonitorConfig *primary_logical_monitor_config = NULL;
-  GfMonitor *primary_monitor;
-  GfLogicalMonitorLayoutMode layout_mode;
-  GList *logical_monitor_configs;
-  GList *region;
-  int x, y;
-  GList *monitors;
-  GList *l;
-
-  primary_monitor = find_primary_monitor (monitor_manager);
-  if (!primary_monitor)
-    return NULL;
-
-  if (!gf_monitor_get_suggested_position (primary_monitor, &x, &y))
-    return NULL;
-
-  layout_mode = gf_monitor_manager_get_default_layout_mode (monitor_manager);
-
-  primary_logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                            primary_monitor,
-                                                                            x, y, NULL,
-                                                                            layout_mode);
-
-  primary_logical_monitor_config->is_primary = TRUE;
-  logical_monitor_configs = g_list_append (NULL, primary_logical_monitor_config);
-  region = g_list_prepend (NULL, &primary_logical_monitor_config->layout);
-
-  monitors = gf_monitor_manager_get_monitors (monitor_manager);
-  for (l = monitors; l; l = l->next)
-    {
-      GfMonitor *monitor = l->data;
-      GfLogicalMonitorConfig *logical_monitor_config;
-
-      if (monitor == primary_monitor)
-        continue;
-
-      if (!gf_monitor_get_suggested_position (monitor, &x, &y))
-        continue;
-
-      logical_monitor_config = create_preferred_logical_monitor_config (monitor_manager,
-                                                                        monitor, x, y,
-                                                                        primary_logical_monitor_config,
-                                                                        layout_mode);
-
-      logical_monitor_configs = g_list_append (logical_monitor_configs, logical_monitor_config);
-
-      if (gf_rectangle_overlaps_with_region (region, &logical_monitor_config->layout))
-        {
-          g_warning ("Suggested monitor config has overlapping region, rejecting");
-          g_list_free (region);
-          g_list_free_full (logical_monitor_configs,
-                            (GDestroyNotify) gf_logical_monitor_config_free);
-
-          return NULL;
-        }
-
-      region = g_list_prepend (region, &logical_monitor_config->layout);
-    }
-
-  g_list_free (region);
-
-  if (!logical_monitor_configs)
-    return NULL;
-
-  return gf_monitors_config_new (monitor_manager, logical_monitor_configs,
-                                 layout_mode, GF_MONITORS_CONFIG_FLAG_NONE);
+  return create_monitors_config (config_manager,
+                                 MONITOR_MATCH_WITH_SUGGESTED_POSITION,
+                                 MONITOR_POSITIONING_SUGGESTED,
+                                 GF_MONITORS_CONFIG_FLAG_NONE);
 }
 
 GfMonitorsConfig *
 gf_monitor_config_manager_create_for_orientation (GfMonitorConfigManager *config_manager,
+                                                  GfMonitorsConfig       *base_config,
                                                   GfMonitorTransform      transform)
 {
-  return create_for_builtin_display_rotation (config_manager, FALSE, transform);
+  return create_for_builtin_display_rotation (config_manager,
+                                              base_config,
+                                              FALSE,
+                                              transform);
+}
+
+GfMonitorsConfig *
+gf_monitor_config_manager_create_for_builtin_orientation (GfMonitorConfigManager *config_manager,
+                                                          GfMonitorsConfig       *base_config)
+{
+  GfMonitorManager *monitor_manager;
+  gboolean panel_orientation_managed;
+  GfMonitorTransform current_transform;
+  GfMonitor *laptop_panel;
+
+  monitor_manager = config_manager->monitor_manager;
+  panel_orientation_managed = gf_monitor_manager_get_panel_orientation_managed (monitor_manager);
+
+  g_return_val_if_fail (panel_orientation_managed, NULL);
+
+  laptop_panel = gf_monitor_manager_get_laptop_panel (monitor_manager);
+  current_transform = get_monitor_transform (monitor_manager, laptop_panel);
+
+  return create_for_builtin_display_rotation (config_manager,
+                                              base_config,
+                                              FALSE,
+                                              current_transform);
 }
 
 GfMonitorsConfig *
 gf_monitor_config_manager_create_for_rotate_monitor (GfMonitorConfigManager *config_manager)
 {
-  return create_for_builtin_display_rotation (config_manager, TRUE, GF_MONITOR_TRANSFORM_NORMAL);
+  if (config_manager->current_config == NULL)
+    return NULL;
+
+  return create_for_builtin_display_rotation (config_manager,
+                                              config_manager->current_config,
+                                              TRUE,
+                                              GF_MONITOR_TRANSFORM_NORMAL);
 }
 
 GfMonitorsConfig *
@@ -1335,7 +1446,21 @@ void
 gf_monitor_config_manager_set_current (GfMonitorConfigManager *config_manager,
                                        GfMonitorsConfig       *config)
 {
-  if (config_manager->current_config)
+  GfMonitorsConfig *current_config;
+  gboolean overrides_current;
+
+  current_config = config_manager->current_config;
+  overrides_current = FALSE;
+
+  if (config != NULL &&
+      current_config != NULL &&
+      has_same_root_config (config, current_config))
+    {
+      overrides_current = gf_monitors_config_key_equal (config->key,
+                                                        current_config->key);
+    }
+
+  if (current_config != NULL && !overrides_current)
     {
       g_queue_push_head (&config_manager->config_history,
                          g_object_ref (config_manager->current_config));
@@ -1380,3 +1505,10 @@ gf_monitor_config_manager_save_current (GfMonitorConfigManager *config_manager)
   gf_monitor_config_store_add (config_manager->config_store,
                                config_manager->current_config);
 }
+
+gboolean
+gf_monitor_manager_is_monitor_visible (GfMonitorManager *monitor_manager,
+                                       GfMonitor        *monitor)
+{
+  return monitor_matches_rule (monitor, monitor_manager, MONITOR_MATCH_VISIBLE);
+}
