@@ -54,6 +54,8 @@ typedef struct
   guint        persistent_timeout_id;
 
   GfPowerSave  power_save_mode;
+
+  gboolean     initial_orient_change_done;
 } GfMonitorManagerPrivate;
 
 typedef gboolean (* MonitorMatchFunc) (GfMonitor *monitor);
@@ -208,10 +210,46 @@ find_monitor (GfMonitorManager *monitor_manager,
   return NULL;
 }
 
-static GfMonitor *
-get_active_monitor (GfMonitorManager *manager)
+static gboolean
+is_global_scale_matching_in_config (GfMonitorsConfig *config,
+                                    float             scale)
 {
-  return find_monitor (manager, gf_monitor_is_active);
+  GList *l;
+
+  for (l = config->logical_monitor_configs; l; l = l->next)
+    {
+      GfLogicalMonitorConfig *logical_monitor_config = l->data;
+
+      if (!G_APPROX_VALUE (logical_monitor_config->scale, scale, FLT_EPSILON))
+        return FALSE;
+    }
+
+  return TRUE;
+}
+
+static gboolean
+is_scale_supported_for_config (GfMonitorManager *self,
+                               GfMonitorsConfig *config,
+                               GfMonitor        *monitor,
+                               GfMonitorMode    *monitor_mode,
+                               float             scale)
+
+{
+  if (gf_monitor_manager_is_scale_supported (self,
+                                             config->layout_mode,
+                                             monitor,
+                                             monitor_mode,
+                                             scale))
+    {
+      if (gf_monitor_manager_get_capabilities (self) &
+          GF_MONITOR_MANAGER_CAPABILITY_GLOBAL_SCALE_REQUIRED)
+        return is_global_scale_matching_in_config (config, scale);
+
+      return TRUE;
+    }
+
+  return FALSE;
+
 }
 
 static gboolean
@@ -219,8 +257,11 @@ gf_monitor_manager_is_config_applicable (GfMonitorManager  *manager,
                                          GfMonitorsConfig  *config,
                                          GError           **error)
 {
+  GfMonitorManagerPrivate *priv;
   GList *l;
 
+  priv = gf_monitor_manager_get_instance_private (manager);
+
   for (l = config->logical_monitor_configs; l; l = l->next)
     {
       GfLogicalMonitorConfig *logical_monitor_config = l->data;
@@ -253,8 +294,11 @@ gf_monitor_manager_is_config_applicable (GfMonitorManager  *manager,
               return FALSE;
             }
 
-          if (!gf_monitor_manager_is_scale_supported (manager, config->layout_mode,
-                                                      monitor, monitor_mode, scale))
+          if (!is_scale_supported_for_config (manager,
+                                              config,
+                                              monitor,
+                                              monitor_mode,
+                                              scale))
             {
               g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
                            "Scale not supported by backend");
@@ -263,7 +307,7 @@ gf_monitor_manager_is_config_applicable (GfMonitorManager  *manager,
             }
 
           if (gf_monitor_is_laptop_panel (monitor) &&
-              gf_monitor_manager_is_lid_closed (manager))
+              gf_backend_is_lid_closed (priv->backend))
             {
               g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
                            "Refusing to activate a closed laptop panel");
@@ -309,11 +353,19 @@ static gfloat
 derive_configured_global_scale (GfMonitorManager *manager,
                                 GfMonitorsConfig *config)
 {
-  GfLogicalMonitorConfig *logical_monitor_config;
+  GList *l;
 
-  logical_monitor_config = config->logical_monitor_configs->data;
+  for (l = config->logical_monitor_configs; l; l = l->next)
+    {
+      GfLogicalMonitorConfig *monitor_config;
 
-  return logical_monitor_config->scale;
+      monitor_config = l->data;
+
+      if (is_global_scale_matching_in_config (config, monitor_config->scale))
+        return monitor_config->scale;
+    }
+
+  return 1.0f;
 }
 
 static gfloat
@@ -323,24 +375,75 @@ calculate_monitor_scale (GfMonitorManager *manager,
   GfMonitorMode *monitor_mode;
 
   monitor_mode = gf_monitor_get_current_mode (monitor);
-  return gf_monitor_manager_calculate_monitor_mode_scale (manager, monitor,
+  return gf_monitor_manager_calculate_monitor_mode_scale (manager,
+                                                          manager->layout_mode,
+                                                          monitor,
                                                           monitor_mode);
 }
 
+static gboolean
+is_scale_supported_by_other_monitors (GfMonitorManager *manager,
+                                      GfMonitor        *not_this_one,
+                                      float             scale)
+{
+  GList *l;
+
+  for (l = manager->monitors; l; l = l->next)
+    {
+      GfMonitor *monitor = l->data;
+      GfMonitorMode *mode;
+
+      if (monitor == not_this_one || !gf_monitor_is_active (monitor))
+        continue;
+
+      mode = gf_monitor_get_current_mode (monitor);
+      if (!gf_monitor_manager_is_scale_supported (manager,
+                                                  manager->layout_mode,
+                                                  monitor,
+                                                  mode,
+                                                  scale))
+        return FALSE;
+    }
+
+  return TRUE;
+}
+
 static gfloat
 derive_calculated_global_scale (GfMonitorManager *manager)
 {
   GfMonitor *monitor;
+  float scale;
+  GList *l;
 
   monitor = gf_monitor_manager_get_primary_monitor (manager);
+  scale = 1.0f;
 
-  if (!monitor || !gf_monitor_is_active (monitor))
-    monitor = get_active_monitor (manager);
+  if (monitor != NULL && gf_monitor_is_active (monitor))
+    {
+      scale = calculate_monitor_scale (manager, monitor);
+      if (is_scale_supported_by_other_monitors (manager, monitor, scale))
+        return scale;
+    }
 
-  if (!monitor)
-    return 1.0;
+  for (l = manager->monitors; l; l = l->next)
+    {
+      GfMonitor *other_monitor;
+      float monitor_scale;
 
-  return calculate_monitor_scale (manager, monitor);
+      other_monitor = l->data;
+
+      if (other_monitor == monitor || !gf_monitor_is_active (other_monitor))
+        continue;
+
+      monitor_scale = calculate_monitor_scale (manager, other_monitor);
+
+      if (is_scale_supported_by_other_monitors (manager,
+                                                other_monitor,
+                                                monitor_scale))
+        scale = MAX (scale, monitor_scale);
+    }
+
+  return scale;
 }
 
 static GfLogicalMonitor *
@@ -498,37 +601,36 @@ gf_monitor_manager_apply_monitors_config (GfMonitorManager        *manager,
 }
 
 static void
-orientation_changed (GfOrientationManager *orientation_manager,
-                     GfMonitorManager     *manager)
+handle_orientation_change (GfOrientationManager *orientation_manager,
+                           GfMonitorManager     *manager)
 {
+  GfOrientation orientation;
   GfMonitorTransform transform;
   GError *error = NULL;
   GfMonitorsConfig *config;
+  GfMonitor *laptop_panel;
+  GfLogicalMonitor *laptop_logical_monitor;
+  GfMonitorsConfig *current_config;
 
-  if (!manager->panel_orientation_managed)
+  laptop_panel = gf_monitor_manager_get_laptop_panel (manager);
+  g_return_if_fail (laptop_panel);
+
+  if (!gf_monitor_is_active (laptop_panel))
     return;
 
-  switch (gf_orientation_manager_get_orientation (orientation_manager))
-    {
-      case GF_ORIENTATION_NORMAL:
-        transform = GF_MONITOR_TRANSFORM_NORMAL;
-        break;
-      case GF_ORIENTATION_BOTTOM_UP:
-        transform = GF_MONITOR_TRANSFORM_180;
-        break;
-      case GF_ORIENTATION_LEFT_UP:
-        transform = GF_MONITOR_TRANSFORM_90;
-        break;
-      case GF_ORIENTATION_RIGHT_UP:
-        transform = GF_MONITOR_TRANSFORM_270;
-        break;
+  orientation = gf_orientation_manager_get_orientation (orientation_manager);
+  transform = gf_monitor_transform_from_orientation (orientation);
 
-      case GF_ORIENTATION_UNDEFINED:
-      default:
-        return;
-    }
+  laptop_logical_monitor = gf_monitor_get_logical_monitor (laptop_panel);
+  if (gf_logical_monitor_get_transform (laptop_logical_monitor) == transform)
+    return;
+
+  current_config = gf_monitor_config_manager_get_current (manager->config_manager);
+  if (current_config == NULL)
+    return;
 
   config = gf_monitor_config_manager_create_for_orientation (manager->config_manager,
+                                                             current_config,
                                                              transform);
 
   if (!config)
@@ -546,6 +648,70 @@ orientation_changed (GfOrientationManager *orientation_manager,
   g_object_unref (config);
 }
 
+/*
+ * Special case for tablets with a native portrait mode and a keyboard dock,
+ * where the device gets docked in landscape mode. For this combo to work
+ * properly with gnome-flashback starting while the tablet is docked, we need
+ * to take the accelerometer reported orientation into account (at
+ * gnome-flashback startup) even if there is a tablet-mode-switch which
+ * indicates that the device is NOT in tablet-mode (because it is docked).
+ */
+static gboolean
+handle_initial_orientation_change (GfOrientationManager *orientation_manager,
+                                   GfMonitorManager     *self)
+{
+  GfMonitor *monitor;
+  GfMonitorMode *mode;
+  int width;
+  int height;
+
+  /*
+   * This is a workaround to ignore the tablet mode switch on the initial
+   * config of devices with a native portrait mode panel. The touchscreen and
+   * accelerometer requirements for applying the orientation must still be met.
+   */
+  if (!gf_orientation_manager_has_accelerometer (orientation_manager))
+    return FALSE;
+
+  /* Check for a portrait mode panel */
+  monitor = gf_monitor_manager_get_laptop_panel (self);
+
+  if (monitor == NULL)
+    return FALSE;
+
+  mode = gf_monitor_get_preferred_mode (monitor);
+  gf_monitor_mode_get_resolution (mode, &width, &height);
+
+  if (width > height)
+    return FALSE;
+
+  handle_orientation_change (orientation_manager, self);
+
+  return TRUE;
+}
+
+static void
+orientation_changed (GfOrientationManager *orientation_manager,
+                     GfMonitorManager     *self)
+{
+  GfMonitorManagerPrivate *priv;
+
+  priv = gf_monitor_manager_get_instance_private (self);
+
+  if (!priv->initial_orient_change_done)
+    {
+      priv->initial_orient_change_done = TRUE;
+
+      if (handle_initial_orientation_change (orientation_manager, self))
+        return;
+    }
+
+  if (!self->panel_orientation_managed)
+    return;
+
+  handle_orientation_change (orientation_manager, self);
+}
+
 static void
 update_panel_orientation_managed (GfMonitorManager *self)
 {
@@ -556,7 +722,8 @@ update_panel_orientation_managed (GfMonitorManager *self)
   priv = gf_monitor_manager_get_instance_private (self);
 
   orientation_manager = gf_backend_get_orientation_manager (priv->backend);
-  panel_orientation_managed = gf_orientation_manager_has_accelerometer (orientation_manager);
+  panel_orientation_managed = gf_orientation_manager_has_accelerometer (orientation_manager) &&
+                              gf_monitor_manager_get_laptop_panel (self);
 
   if (self->panel_orientation_managed == panel_orientation_managed)
     return;
@@ -568,6 +735,10 @@ update_panel_orientation_managed (GfMonitorManager *self)
 
   gf_dbus_display_config_set_panel_orientation_managed (self->display_config,
                                                         panel_orientation_managed);
+
+  /* The orientation may have changed while it was unmanaged */
+  if (panel_orientation_managed)
+    handle_orientation_change (orientation_manager, self);
 }
 
 static void
@@ -582,6 +753,20 @@ restore_previous_config (GfMonitorManager *manager)
       GfMonitorsConfigMethod method;
       GError *error;
 
+      if (manager->panel_orientation_managed)
+        {
+          GfMonitorsConfig *oriented_config;
+
+          oriented_config = gf_monitor_config_manager_create_for_builtin_orientation (manager->config_manager,
+                                                                                      previous_config);
+
+          if (oriented_config != NULL)
+            {
+              g_object_unref (previous_config);
+              previous_config = oriented_config;
+            }
+        }
+
       method = GF_MONITORS_CONFIG_METHOD_TEMPORARY;
       error = NULL;
 
@@ -1075,6 +1260,8 @@ rebuild_monitors (GfMonitorManager *manager)
             }
         }
     }
+
+  update_panel_orientation_managed (manager);
 }
 
 static GList *
@@ -1628,6 +1815,7 @@ gf_monitor_manager_handle_get_current_state (GfDBusDisplayConfig   *skeleton,
 
           preferred_scale =
             gf_monitor_manager_calculate_monitor_mode_scale (manager,
+                                                             manager->layout_mode,
                                                              monitor,
                                                              monitor_mode);
 
@@ -2008,15 +2196,6 @@ gf_monitor_manager_real_read_edid (GfMonitorManager *manager,
   return NULL;
 }
 
-static gboolean
-gf_monitor_manager_real_is_lid_closed (GfMonitorManager *manager)
-{
-  if (!manager->up_client)
-    return FALSE;
-
-  return manager->lid_is_closed;
-}
-
 static void
 gf_monitor_manager_real_read_current_state (GfMonitorManager *manager)
 {
@@ -2043,33 +2222,21 @@ gf_monitor_manager_real_read_current_state (GfMonitorManager *manager)
 }
 
 static void
-lid_is_closed_changed (UpClient   *client,
-                       GParamSpec *pspec,
-                       gpointer    user_data)
+lid_is_closed_changed_cb (GfBackend        *backend,
+                          GParamSpec       *pspec,
+                          GfMonitorManager *self)
 {
-  GfMonitorManager *manager;
-  gboolean lid_is_closed;
-
-  manager = user_data;
-  lid_is_closed = up_client_get_lid_is_closed (manager->up_client);
-
-  if (lid_is_closed == manager->lid_is_closed)
-    return;
-
-  manager->lid_is_closed = lid_is_closed;
-  gf_monitor_manager_ensure_configured (manager);
+  gf_monitor_manager_ensure_configured (self);
 }
 
 static void
 gf_monitor_manager_constructed (GObject *object)
 {
   GfMonitorManager *manager;
-  GfMonitorManagerClass *manager_class;
   GfMonitorManagerPrivate *priv;
   GfOrientationManager *orientation_manager;
 
   manager = GF_MONITOR_MANAGER (object);
-  manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
   priv = gf_monitor_manager_get_instance_private (manager);
 
   G_OBJECT_CLASS (gf_monitor_manager_parent_class)->constructed (object);
@@ -2077,19 +2244,6 @@ gf_monitor_manager_constructed (GObject *object)
   manager->display_config = gf_dbus_display_config_skeleton_new ();
   monitor_manager_setup_dbus_config_handlers (manager);
 
-  if (manager_class->is_lid_closed == gf_monitor_manager_real_is_lid_closed)
-    {
-      manager->up_client = up_client_new ();
-
-      if (manager->up_client)
-        {
-          g_signal_connect_object (manager->up_client, "notify::lid-is-closed",
-                                   G_CALLBACK (lid_is_closed_changed), manager, 0);
-
-          manager->lid_is_closed = up_client_get_lid_is_closed (manager->up_client);
-        }
-    }
-
   g_signal_connect_object (manager->display_config, "notify::power-save-mode",
                            G_CALLBACK (power_save_mode_changed), manager,
                            G_CONNECT_SWAPPED);
@@ -2104,6 +2258,12 @@ gf_monitor_manager_constructed (GObject *object)
                            manager,
                            G_CONNECT_SWAPPED);
 
+  g_signal_connect_object (priv->backend,
+                           "lid-is-closed-changed",
+                           G_CALLBACK (lid_is_closed_changed_cb),
+                           manager,
+                           0);
+
   manager->current_switch_config = GF_MONITOR_SWITCH_CONFIG_UNKNOWN;
 
   priv->bus_name_id = g_bus_own_name (G_BUS_TYPE_SESSION,
@@ -2132,9 +2292,14 @@ gf_monitor_manager_dispose (GObject *object)
       priv->bus_name_id = 0;
     }
 
+  if (priv->persistent_timeout_id != 0)
+    {
+      g_source_remove (priv->persistent_timeout_id);
+      priv->persistent_timeout_id = 0;
+    }
+
   g_clear_object (&manager->display_config);
   g_clear_object (&manager->config_manager);
-  g_clear_object (&manager->up_client);
 
   priv->backend = NULL;
 
@@ -2274,7 +2439,6 @@ gf_monitor_manager_class_init (GfMonitorManagerClass *manager_class)
   object_class->set_property = gf_monitor_manager_set_property;
 
   manager_class->read_edid = gf_monitor_manager_real_read_edid;
-  manager_class->is_lid_closed = gf_monitor_manager_real_is_lid_closed;
   manager_class->read_current_state = gf_monitor_manager_real_read_current_state;
 
   gf_monitor_manager_install_properties (object_class);
@@ -2413,9 +2577,9 @@ gf_monitor_manager_read_current_state (GfMonitorManager *manager)
 }
 
 void
-gf_monitor_manager_on_hotplug (GfMonitorManager *manager)
+gf_monitor_manager_reconfigure (GfMonitorManager *self)
 {
-  gf_monitor_manager_ensure_configured (manager);
+  gf_monitor_manager_ensure_configured (self);
 }
 
 gboolean
@@ -2499,10 +2663,25 @@ gf_monitor_manager_ensure_configured (GfMonitorManager *manager)
 
       if (config)
         {
+          GfMonitorsConfig *oriented_config;
+
+          oriented_config = NULL;
+
+          if (manager->panel_orientation_managed)
+            {
+              oriented_config = gf_monitor_config_manager_create_for_builtin_orientation (manager->config_manager,
+                                                                                          config);
+
+              if (oriented_config != NULL)
+                config = oriented_config;
+            }
+
           if (!gf_monitor_manager_apply_monitors_config (manager, config,
                                                          method, &error))
             {
               config = NULL;
+              g_clear_object (&oriented_config);
+
               g_warning ("Failed to use stored monitor configuration: %s",
                          error->message);
               g_clear_error (&error);
@@ -2510,6 +2689,41 @@ gf_monitor_manager_ensure_configured (GfMonitorManager *manager)
           else
             {
               g_object_ref (config);
+              g_clear_object (&oriented_config);
+              goto done;
+            }
+        }
+    }
+
+  if (manager->panel_orientation_managed)
+    {
+      GfMonitorsConfig *current_config;
+
+      current_config = gf_monitor_config_manager_get_current (manager->config_manager);
+
+      if (current_config != NULL)
+        {
+          config = gf_monitor_config_manager_create_for_builtin_orientation (manager->config_manager,
+                                                                             current_config);
+        }
+    }
+
+  if (config != NULL)
+    {
+      if (gf_monitor_manager_is_config_complete (manager, config))
+        {
+          if (!gf_monitor_manager_apply_monitors_config (manager,
+                                                         config,
+                                                         method,
+                                                         &error))
+            {
+              g_clear_object (&config);
+              g_warning ("Failed to use current monitor configuration: %s",
+                         error->message);
+              g_clear_error (&error);
+            }
+          else
+            {
               goto done;
             }
         }
@@ -2535,7 +2749,21 @@ gf_monitor_manager_ensure_configured (GfMonitorManager *manager)
   config = gf_monitor_config_manager_get_previous (manager->config_manager);
   if (config)
     {
+      GfMonitorsConfig *oriented_config;
+
+      oriented_config = NULL;
+
+      if (manager->panel_orientation_managed)
+        {
+          oriented_config = gf_monitor_config_manager_create_for_builtin_orientation (manager->config_manager,
+                                                                                      config);
+
+          if (oriented_config != NULL)
+            config = oriented_config;
+        }
+
       config = g_object_ref (config);
+      g_clear_object (&oriented_config);
 
       if (gf_monitor_manager_is_config_complete (manager, config))
         {
@@ -2624,22 +2852,20 @@ gf_monitor_manager_update_logical_state_derived (GfMonitorManager *manager,
   gf_monitor_manager_rebuild_logical_monitors_derived (manager, config);
 }
 
-gboolean
-gf_monitor_manager_is_lid_closed (GfMonitorManager *manager)
-{
-  return GF_MONITOR_MANAGER_GET_CLASS (manager)->is_lid_closed (manager);
-}
-
 gfloat
-gf_monitor_manager_calculate_monitor_mode_scale (GfMonitorManager *manager,
-                                                 GfMonitor        *monitor,
-                                                 GfMonitorMode    *monitor_mode)
+gf_monitor_manager_calculate_monitor_mode_scale (GfMonitorManager           *manager,
+                                                 GfLogicalMonitorLayoutMode  layout_mode,
+                                                 GfMonitor                  *monitor,
+                                                 GfMonitorMode              *monitor_mode)
 {
   GfMonitorManagerClass *manager_class;
 
   manager_class = GF_MONITOR_MANAGER_GET_CLASS (manager);
 
-  return manager_class->calculate_monitor_mode_scale (manager, monitor, monitor_mode);
+  return manager_class->calculate_monitor_mode_scale (manager,
+                                                      layout_mode,
+                                                      monitor,
+                                                      monitor_mode);
 }
 
 gfloat *
@@ -2800,7 +3026,11 @@ gf_monitor_manager_get_switch_config (GfMonitorManager *manager)
 gboolean
 gf_monitor_manager_can_switch_config (GfMonitorManager *manager)
 {
-  return (!gf_monitor_manager_is_lid_closed (manager) &&
+  GfMonitorManagerPrivate *priv;
+
+  priv = gf_monitor_manager_get_instance_private (manager);
+
+  return (!gf_backend_is_lid_closed (priv->backend) &&
           g_list_length (manager->monitors) > 1);
 }
 
